#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HEX_MAX_SIZE  (1024*8*3)
#define BIN_MAX_SIZE  (1024*8)
#define SEC_MAX_SIZE  (64)
#define LISP_MAX_SIZE (1024)

unsigned short crc16_hex;

int hex1(unsigned char ch) {
	if (ch >= '0' && ch <= '9')
		return ch - '0';
	else if (ch >= 'A' && ch <= 'F')
		return ch - 'A' + 10;
	else if (ch >= 'a' && ch <= 'f')
		return ch - 'a' + 10;
	else
		return -1;
}

int str_to_hex(unsigned char *str, int len, uint32_t *pval) {
	int offset = 0;
	int tmp;

	*pval = 0;
	
	switch (len) {
	case 8:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 7:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 6:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 5:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 4:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 3:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 2:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		*pval <<= 4;
	case 1:
		tmp = hex1(str[offset++]);
		if (tmp == -1)
			return 2;
		*pval |= tmp;
		return 0;
	case 0:
		return 1;
	default:
		return 1;
	};
}

int crc_step(unsigned char *str, int old_crc) {
	uint32_t val;
	int ret;
	
	ret = str_to_hex(str, 2, &val);
	if (ret)
		return -1;
	
	return (old_crc + val) & 0xFF;
}

unsigned short crc16_cy(unsigned char *block, unsigned short len) {
	unsigned short crc = 0x0000;
	while (len--)
		crc += *block++;
	return crc;
}

int hex_load(unsigned char *data, long size, unsigned char *pbinbuff, long binbuff_len, unsigned char *psecbuff, long secbuff_len) {
	int ret;
	int i;
	long offset = 0;
	int mode = 0;
	
	crc16_hex = 0;
	
	while (offset < size) {
		if (data[offset] == ':') {
			uint32_t len;
			uint32_t val;
			uint32_t addr;
			uint32_t type;
			int crc = 0;
			offset++;
			ret = str_to_hex(&data[offset], 2, &len);
			if (ret)
				return 2;
			ret = str_to_hex(&data[offset+2], 4, &addr);
			if (ret)
				return 2;
			ret = str_to_hex(&data[offset+6], 2, &type);
			if (ret)
				return 2;
			crc = crc_step(&data[offset], crc);
			crc = crc_step(&data[offset+2], crc);
			crc = crc_step(&data[offset+4], crc);
			crc = crc_step(&data[offset+6], crc);
			offset += 8;
			
			for (i = 0; i < len; i++) {
				ret = str_to_hex(&data[offset], 2, &val);
				if (ret)
					return 3;
				
				crc = crc_step(&data[offset], crc);
				offset += 2;
				
				if (type == 0x00) {
					if (mode == 0) {
						if (addr+i > binbuff_len) {
							printf("ERROR: Too big HEX code!\r\n");
							return 6;
						}
						pbinbuff[addr+i] = val;
					} else 
					if (mode == 1) {
						if (addr+i > secbuff_len) {
							printf("ERROR: Too many secure data in HEX!\r\n");
							return 6;
						}
						psecbuff[addr+i] = val;
					} else 
					if (mode == 2) {
						if (i > 1) {
							printf("ERROR: Too many CRC data in HEX!\r\n");
							return 6;
						}
						crc16_hex <<= 8;
						crc16_hex |= val;
					}
				} else 
				if (type == 0x04) {
					if (i == 0) {
						if (val == 0x00) {
							continue;
						} else {
							printf("ERROR: Unknown HEX section value (1)!\r\n");
							return 7;
						}
					}
					if (val == 0x10 && i == 1) {
						mode = 1;
					} else 
					if (val == 0x20 && i == 1) {
						mode = 2;
					} else {
						printf("ERROR: Unknown HEX section value (1)!\r\n");
					}
				} else {
					printf("ERROR: Unknown HEX section 0x%02x!\r\n", type);
					return 8;
				}
			}
			crc = crc_step(&data[offset], crc);
			offset += 2;
			if (crc != 0) {
				return 4;
			}
			if (data[offset] == '\r') {
				offset++;
			}
			if (data[offset] != '\n') {
				return 5;
			}
			offset++;
		} else {
			return 1;
		}
	}
	return 0;
}

int memfind(unsigned char *bin, int size, unsigned char *find, int fsize) {
	int offset = 0;
	int foffset = 0;
	while (offset+foffset < size) {
		if (bin[offset+foffset] == find[foffset]) {
			foffset++;
			if (foffset == fsize) {
				return offset;
			}
		} else {
			foffset = 0;
			offset++;
		}
	}
	return -1;
}

int main(int argc, char **argv) {
	FILE *f;
	int ret;
	int size_hex = 0;
	int size_lisp = 0;
	unsigned char program_hex[HEX_MAX_SIZE];
	unsigned char program_bin[BIN_MAX_SIZE];
	unsigned char program_lisp[LISP_MAX_SIZE];
	unsigned char program_sec[SEC_MAX_SIZE];
	unsigned char mem[4];
	int sub_start;
	int sub_end;
	unsigned short crc16;
	// Help:
	if (argc != 3) {
		printf("uLisp write to HEX\r\n\r\n");
		printf("ulisp_write.exe file.hex lisp.ls\r\n");
		return 0;
	}
	// Load firmware HEX file:
	f = fopen(argv[1],"r");
	if (f == NULL) {
		printf("error: bad input file \'%s\'!\r\n", argv[1]);
		return 1;
	}
	while (1) {
		int v = fgetc(f);
		if (v == EOF) {
			break;
		} else if (size_hex == sizeof(program_hex)) {
			printf("ERROR: too big program HEX!\r\n");
			return 1;
		} else {
			program_hex[size_hex] = v;
			size_hex++;
		}
	}
	fclose (f);
	// Load lisp file:
	f = fopen(argv[2],"r");
	if (f == NULL) {
		printf("error: bad input file \'%s\'!\r\n", argv[2]);
		return 1;
	}
	while (1) {
		int v = fgetc(f);
		if (v == EOF) {
			break;
		} else if (size_lisp == sizeof(program_lisp)) {
			printf("ERROR: too big Lisp program!\r\n");
			return 1;
		} else {
			program_lisp[size_lisp] = v;
			size_lisp++;
		}
	}
	fclose (f);	
	// Hex
	//printf("%s", program_hex);
	ret = hex_load(program_hex, size_hex, program_bin, sizeof(program_bin), program_sec, sizeof(program_sec));
	if (ret) {
		printf("ERROR: hex_load() return %d!\r\n", ret);
		return 1;
	}
	// check CRC in HEX file:
	crc16 = crc16_cy(program_bin, sizeof(program_bin));
	if (crc16 != crc16_hex) {
		printf("ERROR: CRC in HEX %04x, but real CRC is %04x!\r\n", crc16_hex, crc16);
		return 1;
	} else {
		printf("Firmware CRC: 0x%04x\r\n", crc16);
	}
	// find memory:
	mem[0] = 0xAA; mem[1] = 0xBB; mem[2] = 0xCC; mem[3] = 0xDD; 
	sub_start = memfind(program_bin, sizeof(program_bin), mem, sizeof(mem));
	if (sub_start == -1) {
		printf("ERROR: start label didn't found in HEX!\r\n");
		return 1;
	}
	sub_start += 4;
	mem[0] = 0xDD; mem[1] = 0xCC; mem[2] = 0xBB; mem[3] = 0xAA; 
	sub_end = memfind(&program_bin[sub_start], sizeof(program_bin), mem, sizeof(mem));
	if (sub_end == -1) {
		printf("ERROR: end label didn't found in HEX!\r\n");
		return 1;
	}
	sub_end += sub_start;
	// check size:
	if ((sub_end - sub_start) != sizeof(program_lisp)) {
		printf("ERROR: incorrect buffer size in HEX program (%d)!\r\n", (sub_end - sub_start));
		return 1;
	}
	// insert Lisp program:
	memcpy(&program_bin[sub_start], program_lisp, size_lisp);
	// save HEX back:
	{
		int i;
		int offset = 0;
		unsigned short crc = 0;
		
		f = fopen("out.hex","wb");
		if (f == NULL) {
			printf("error: cannot create output file!\r\n");
			return 1;
		}
		// BIN
		while (offset < sizeof(program_bin)) {
			crc  = 0x40;
			crc += ((offset&0xFF00)>>8) + (offset&0xFF);
			fprintf(f, ":40%04x00", offset);
			for (i=0; i<64; i++) {
				fprintf(f, "%02x", program_bin[offset+i]);
				crc += program_bin[offset+i];
			}
			fprintf(f, "%02x\r\n", (0x100 - (crc&0xFF))&0xFF);
			offset += i;
		}
		// SEC
		fprintf(f, ":020000040010ea\r\n");
		fprintf(f, ":40000000");
		crc = 0x40;
		for (i = 0; i<sizeof(program_sec); i++) {
			fprintf(f, "%02x", program_sec[i]);
			crc += program_sec[i];
		}
		fprintf(f, "%02x\r\n", (0x100 - (crc&0xFF))&0xFF);
		// CRC16
		fprintf(f, ":020000040020da\r\n");
		crc16_hex = crc16_cy(program_bin, sizeof(program_bin));
		fprintf(f, ":02000000%04x%02x\r\n", crc16_hex, (0x100-(0x02+(crc16_hex&0xFF)+(crc16_hex>>8)))&0xFF);
		// END
		fprintf(f, ":00000001ff\r\n");
		fclose (f);	
	}
	return 0;
}
